/*
  ProgEx35 -- Random Binary File Access

  For this lets start with the Employee struct we've been working with for some
  time now.  Up to now I havn't discussed the fopen() stdio.h function very 
  much, but now its time to look at it.  The first parameter is a char* param
  that constitutes the path of the file to open.  The second parameter is termed
  mode and its the tricky one.  In fact its so tricky that over the years I've
  gradually moved away from using the C runtime file functions and I've 
  gravitated to the Win32 Api functions.  However, the C runtime functions are
  highly usable so I want to cover them.  I've decided to just list Microsoft's
  description of their fopen() mode arguement...

  The character string mode specifies the type of access requested for the file, 
  as follows:

  "r"  Opens for reading. If the file does not exist or cannot be found, the 
       fopen call fails.

  "w"  Opens an empty file for writing. If the given file exists, its contents 
       are destroyed.

  "a"  Opens for writing at the end of the file (appending) without removing the 
       EOF marker before writing new data to the file; creates the file first if 
       it doesn't exist.

  "r+" Opens for both reading and writing. (The file must exist.)

  "w+" Opens an empty file for both reading and writing. If the given file 
       exists, its contents are destroyed.

  "a+" Opens for reading and appending; the appending operation includes the 
       removal of the EOF marker before new data is written to the file and the 
       EOF marker is restored after writing is complete; creates the file first 
       if it doesn't exist.  When a file is opened with the "a" or "a+" access 
       type, all write operations occur at the end of the file. The file pointer 
       can be repositioned using fseek or rewind, but is always moved back to 
       the end of the file before any write operation is carried out. Thus, 
       existing data cannot be overwritten.

       The "a" mode does not remove the EOF marker before appending to the file. 
       After appending has occurred, the MS-DOS TYPE command only shows data up 
       to the original EOF marker and not any data appended to the file. The 
       "a+" mode does remove the EOF marker before appending to the file. After 
       appending, the MS-DOS TYPE command shows all data in the file. The "a+" 
       mode is required for appending to a stream file that is terminated with 
       the CTRL+Z EOF marker.

       When the "r+", "w+", or "a+" access type is specified, both reading and 
       writing are allowed (the file is said to be open for "update"). However, 
       when you switch between reading and writing, there must be an 
       intervening fflush, fsetpos, fseek, or rewind operation. The current 
       position can be specified for the fsetpos or fseek operation, if desired.

                               --End Microsoft--

       Lets start with a simple program to just write those two Employee records
       we've been playing with all along to a random access file.  That will 
       give you a view of the fwrite() and fread() functions.  These are 
       something like BASIC's Get and Put statements.  They are prototyped like 
       so...

       size_t fwrite(const void* buffer, size_t size, size_t count, FILE* stream);
       size_t fread(void* buffer, size_t size, size_t count, FILE* stream);

       Parameters

       buffer      Storage location for data
       size        Item size in bytes
       count       Maximum number of items to be written/read
       stream      Pointer to FILE structure

       size_t is a 4 byte quantity on 32 bit systems.  To start, we'll assume 
       we're creating a new file that does not previusly exist, so our fopen() 
       will look like this...

       fp=fopen("Data.dat","wb");

       I've used "wb" for mode.  The "w" is to write of course, and the "b"
       means binary mode.  With this mode setup any previously existing Data.dat
       would be destroyed.  

       The fwrite() function can easily write 1 or more records to the file.  In
       the program below we have this call pertaining to our 2 Employee records
       emp[2]...

       iCount = fwrite(emp, sizeof(Employee), 2, fp); 

       Note that the 1st parameter of the function takes a pointer, and I've
       said many times here that an array name without the brackets is a base
       pointer to the array, so you see we just use 'emp'.  If our declaration
       of the Employee struct/type was just...

       Employee emp;

       and we were just writting that one record, the call would look like 
       this...

       iCount = fwrite(&emp, sizeof(Employee), 1, fp);

       The iCount variable just receives the number of records written, so it is
       useful for error testing.
*/
#include <stdio.h>
#include <string.h>

struct Employee          //Same as Type Employee
{
 char szName[32];   
 char szAddress[32];
 char szCity[16];
 char szState[16];
 char szCountry[16];
 char szZipCode[16];
 unsigned int iAge;
 unsigned int iEmpNum;
 double dblSalary;   
};                       //End Type

void Initialize(Employee* pEmp)
{
 strcpy(pEmp[0].szName,"Martin Veneski");          //1st employee
 strcpy(pEmp[0].szAddress,"1016 Wabash Street");
 strcpy(pEmp[0].szCity,"Shamokin");
 strcpy(pEmp[0].szState,"PA");
 strcpy(pEmp[0].szCountry,"USA");
 strcpy(pEmp[0].szZipCode,"17872");
 pEmp[0].iAge=23;
 pEmp[0].iEmpNum=123456;
 pEmp[0].dblSalary=45000;
 
 strcpy(pEmp[1].szName,"Ray Latchkoski");         //2nd employee
 strcpy(pEmp[1].szAddress,"1624 Tioga Street");
 strcpy(pEmp[1].szCity,"Shamokin");
 strcpy(pEmp[1].szState,"PA");
 strcpy(pEmp[1].szCountry,"USA");
 strcpy(pEmp[1].szZipCode,"17872");
 pEmp[1].iAge=25;
 pEmp[1].iEmpNum=789123;
 pEmp[1].dblSalary=50000;
}

int main(void)
{
 Employee emp[2];
 FILE* fp=NULL;
 int iCount=0;

 Initialize(emp);
 fp=fopen("Data.dat","wb");                   //"wb" for a mode parameter will
 if(fp)                                       //open the file in binary mode and
 {                                            //destroy any existing contents of
    puts("File Apparently Opened!");          //a file named Data.dat.  The 3rd
    iCount=fwrite(emp,sizeof(Employee),2,fp); //parameter of the fwrite call is
    printf("iCount = %u\n",iCount);           //2 so we're writting both emp[]
    fclose(fp);                               //records to the file.
 }
 else
 {
    puts("Couldn't Open File!");
    return -1;
 }

 memset(emp,0,sizeof(Employee)*2);            //Erase the existing data in emp[]
 fp=fopen("Data.dat","rb");                   //with memset().  Then use "rb" or
 if(fp)                                       //'read binary' as a mode param.
 {                                            //If the file Data.dat doesn't
    puts("File Apparently Opened!");          //exist the fopen() call will 
    iCount=fread(emp,sizeof(Employee),2,fp);  //fail.  We'll read both records
    printf("iCount = %u\n\n",iCount);         //back into emp[] with fread(), 
    fclose(fp);                               //but we'll need a loop for output
    puts("Rec#    Name            Address                 City            State   Country ZipCode Age     EmpNum  Salary");
    puts("=================================================================================================================");
    for(unsigned int i=0; i<2; i++)
    {
        printf
        (
         "%u\t%s\t%s\t%s\t%s\t%s\t%s\t%u\t%u\t%6.2f\n",
         i,
         emp[i].szName,
         emp[i].szAddress,
         emp[i].szCity,
         emp[i].szState,
         emp[i].szCountry,
         emp[i].szZipCode,
         emp[i].iAge,
         emp[i].iEmpNum,
         emp[i].dblSalary
        );
    }
 }
 else
    puts("Couldn't Open File!");
 getchar();

 return 0;
}

/*
File Apparently Opened!
iCount = 2
File Apparently Opened!
iCount = 2

Rec#    Name            Address                 City            State   Country ZipCode Age     EmpNum  Salary
=================================================================================================================
0       Martin Veneski  1016 Wabash Street      Shamokin        PA      USA     17872   23      123456  45000.00
1       Ray Latchkoski  1624 Tioga Street       Shamokin        PA      USA     17872   25      789123  50000.00
*/
